{-|
Module      : Gargantext.Core.Worker.Env
Description : Asynchronous worker logic (environment)
Copyright   : (c) CNRS, 2024
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

-}

{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# OPTIONS_GHC -Wno-orphans      #-}  -- orphan HasNodeError IOException


module Gargantext.Core.Worker.Env where


import Control.Lens (prism', to)
import Control.Monad.Trans.Control (MonadBaseControl)
import Data.Pool (Pool)
import Data.Text qualified as T
import Database.PostgreSQL.Simple (Connection)
import Gargantext.API.Admin.EnvTypes (Mode(Dev), modeToLoggingLevels)
import Gargantext.API.Admin.Settings ( devJwkFile, devSettings, newPool, SettingsFile (..), IniFile (..) )
import Gargantext.API.Admin.Types (HasSettings(..), Settings(..))
import Gargantext.API.Prelude (GargM)
import Gargantext.Core.Config (GargConfig(..), readConfig)
import Gargantext.Core.Config.Mail qualified as Mail
import Gargantext.Core.Config.NLP qualified as NLP
import Gargantext.Core.Mail.Types (HasMail(..))
import Gargantext.Core.NLP (HasNLPServer(..), NLPServerMap, nlpServerMap)
import Gargantext.Core.NodeStory (NodeStoryEnv, fromDBNodeStoryEnv)
import Gargantext.Database.Prelude (HasConfig(..), HasConnectionPool(..), databaseParameters)
import Gargantext.Database.Query.Table.Node.Error (HasNodeError(..))
import Gargantext.Prelude hiding (to)
import Gargantext.System.Logging (HasLogger(..), Logger, withLoggerHoisted)
import Prelude qualified
import System.Log.FastLogger qualified as FL


data WorkerEnv = WorkerEnv
  { _w_env_settings  :: !Settings
  , _w_env_config    :: !GargConfig
  , _w_env_logger    :: !(Logger (GargM WorkerEnv IOException))
  , _w_env_pool      :: !(Pool Connection)
  , _w_env_nodeStory :: !NodeStoryEnv
  , _w_env_mail      :: !Mail.MailConfig
  , _w_env_nlp       :: !NLPServerMap
  }


withWorkerEnv :: IniFile -> SettingsFile -> (WorkerEnv -> IO a) -> IO a
withWorkerEnv (IniFile iniPath) settingsFile k = withLoggerHoisted Dev $ \logger -> do
  env <- newWorkerEnv logger
  k env -- `finally` cleanEnv env

  where
    newWorkerEnv logger = do
      cfg     <- readConfig         iniPath
      dbParam <- databaseParameters iniPath
      --nodeStory_env <- fromDBNodeStoryEnv (_gc_repofilepath cfg)
      pool    <- newPool            dbParam
      nodeStory_env <- fromDBNodeStoryEnv pool
      setts   <- devSettings devJwkFile settingsFile
      mail    <- Mail.readConfig iniPath
      nlp_config <- NLP.readConfig iniPath
      pure $ WorkerEnv
        { _w_env_pool      = pool
        , _w_env_logger    = logger
        , _w_env_nodeStory = nodeStory_env
        , _w_env_settings  = setts
        , _w_env_config    = cfg
        , _w_env_mail      = mail
        , _w_env_nlp       = nlpServerMap nlp_config
        }

instance HasConfig WorkerEnv where
  hasConfig = to _w_env_config

instance HasSettings WorkerEnv where
  settings = to _w_env_settings

instance HasLogger (GargM WorkerEnv IOException) where
  data instance Logger (GargM WorkerEnv IOException)  =
    GargWorkerLogger {
        w_logger_mode    :: Mode
      , w_logger_set     :: FL.LoggerSet
      }
  type instance LogInitParams (GargM WorkerEnv IOException) = Mode
  type instance LogPayload (GargM WorkerEnv IOException)    = FL.LogStr
  initLogger                = \mode -> do
    w_logger_set <- liftIO $ FL.newStderrLoggerSet FL.defaultBufSize
    pure $ GargWorkerLogger mode w_logger_set
  destroyLogger = \GargWorkerLogger{..}  -> liftIO $ FL.rmLoggerSet w_logger_set
  logMsg = \(GargWorkerLogger mode logger_set) lvl msg -> do
    let pfx = "[" <> show lvl <> "] " :: Text
    when (lvl `elem` (modeToLoggingLevels mode)) $
      liftIO $ FL.pushLogStrLn logger_set $ FL.toLogStr pfx <> msg
  logTxt lgr lvl msg = logMsg lgr lvl (FL.toLogStr $ T.unpack msg)

instance HasConnectionPool WorkerEnv where
  connPool = to _w_env_pool

instance HasMail WorkerEnv where
  mailSettings = to _w_env_mail

instance HasNLPServer WorkerEnv where
  nlpServer = to _w_env_nlp

---------------

newtype WorkerMonad a =
  WorkerMonad { _WorkerMonad :: GargM WorkerEnv IOException a }
  deriving ( Functor
           , Applicative
           , Monad
           , MonadIO
           , MonadReader WorkerEnv
           , MonadBase IO
           , MonadBaseControl IO
           , MonadError IOException
           , MonadFail )
instance HasNodeError IOException where
  _NodeError = prism' (Prelude.userError . show) (const Nothing)

runWorkerMonad :: WorkerEnv -> WorkerMonad a -> IO a
runWorkerMonad env m = do
  res <- runExceptT . flip runReaderT env $ _WorkerMonad m
  case res of
    Left e -> throwIO e
    Right x -> pure x
